"""
Module containing various decorators
"""

from __future__ import annotations
from inspect import iscoroutinefunction

from typing import (
    cast,
    Any,
    Awaitable,
    Concatenate,
    Callable,
    Protocol,
    ParamSpec,
    TypeVar,
    TYPE_CHECKING,
    overload
)

from slixmpp import JID
from poezio import common

if TYPE_CHECKING:
    from poezio.core.core import Core


P = ParamSpec('P')
C = TypeVar('C', bound=Callable[..., Any])
T = TypeVar('T')
U = TypeVar('U')
V = TypeVar('V')


class HasCore(Protocol):
    core: 'Core'


CoreSelf = TypeVar('CoreSelf', bound=HasCore)


def wrap_before_quoted(
        func: Callable[Concatenate[T, list[str], P], U],
        before: Callable[[str], list[str]]
) -> Callable[Concatenate[T, str, P], U]:
    """
    Wrapper for functions requiring "before" handlers.
    """

    if iscoroutinefunction(func):
        def awrap(self: T, args: str, /, *list_args: P.args, **kwargs: P.kwargs) -> U:
            new_args = before(args)
            return cast(U, func(self, new_args, *list_args, **kwargs))
        return awrap
    else:
        def wrap(self: T, args: str, /, *list_args: P.args, **kwargs: P.kwargs) -> U:
            new_args = before(args)
            return func(self, new_args, *list_args, **kwargs)
        return wrap


def wrap_before_ignored(
        func: Callable[Concatenate[T, P], U],
) -> Callable[Concatenate[T, str, P], U]:
    """
    Wrapper for functions requiring "before" handlers.
    """

    if iscoroutinefunction(func):
        def awrap(self: T, args: str, /, *list_args: P.args, **kwargs: P.kwargs) -> U:
            return cast(U, func(self, *list_args, **kwargs))
        return awrap
    else:
        def wrap(self: T, args: str, /, *list_args: P.args, **kwargs: P.kwargs) -> U:
            return func(self, *list_args, **kwargs)
        return wrap


def wrap_before_cancel_on_false(
        func: Callable[P, U],
        before: Callable[[list[Any], dict[str, Any]], bool],
) -> Callable[P, U | bool]:
    """
    Wrapper for functions requiring "before" handlers.
    """

    if iscoroutinefunction(func):
        def awrap(*list_args: P.args, **kwargs: P.kwargs) -> U | bool:
            if before(cast(list[Any], list_args), cast(dict[str, Any], kwargs)):
                return cast(U, func(*list_args, **kwargs))
            return False
        return awrap
    else:
        def wrap(*list_args: P.args, **kwargs: P.kwargs) -> U | bool:
            if before(cast(list[Any], list_args), cast(dict[str, Any], kwargs)):
                return func(*list_args, **kwargs)
            return False
        return wrap


@overload
def wrap_after(
        func: Callable[P, Awaitable[V]],
        after: Callable[[U, list[Any], dict[str, Any]], None],
) -> Callable[P, Awaitable[V]]:
    ...


@overload
def wrap_after(
        func: Callable[P, U],
        after: Callable[[U, list[Any], dict[str, Any]], None],
) -> Callable[P, U]:
    ...


def wrap_after(
        func: Callable[P, Awaitable[V]] | Callable[P, U],
        after: Callable[[U, list[Any], dict[str, Any]], None],
) -> Callable[P, Awaitable[V]] | Callable[P,  U]:
    """
    Wrapper for functions requiring "before" handlers.
    """

    if iscoroutinefunction(func):
        async def awrap(*list_args: P.args, **kwargs: P.kwargs) -> V:
            result = await func(*list_args, **kwargs)
            after(result, cast(list[Any], list_args), cast(dict[str, Any], kwargs))
            return cast(V, result)
        return awrap
    else:
        def wrap(*list_args: P.args, **kwargs: P.kwargs) -> U:
            result = cast(U, func(*list_args, **kwargs))
            after(result, cast(list[Any], list_args), cast(dict[str, Any], kwargs))
            return result
        return wrap


class RefreshWrapper:
    core: 'Core | None'

    def __init__(self) -> None:
        self.core = None

    def conditional(self, func: C) -> C:
        """
        Decorator to refresh the UI if the wrapped function
        returns True
        """
        def after(result: Any, args: list[Any], kwargs: dict[str, Any]) -> Any:
            if self.core is not None and result:
                self.core.refresh_window()  # pylint: disable=no-member
            return result

        wrap = wrap_after(func, after=after)

        return cast(C, wrap)

    def always(self, func: C) -> C:
        """
        Decorator that refreshes the UI no matter what after the function
        """
        def after(result: Any, args: list[Any], kwargs: dict[str, Any]) -> Any:
            if self.core is not None:
                self.core.refresh_window()  # pylint: disable=no-member
            return result

        wrap = wrap_after(func, after=after)
        return cast(C, wrap)

    def update(self, func: C) -> C:
        """
        Decorator that only updates the screen
        """

        def after(result: Any, args: list[Any], kwargs: dict[str, Any]) -> Any:
            if self.core is not None:
                self.core.doupdate()  # pylint: disable=no-member
            return result
        wrap = wrap_after(func, after=after)
        return cast(C, wrap)


refresh_wrapper = RefreshWrapper()


class CommandArgParser:
    """Modify the string argument of the function into a list of strings
    containing the right number of extracted arguments, or None if we don’t
    have enough.
    """

    @staticmethod
    def raw(func: T) -> T:
        """Just call the function with a single string, which is the original string
        untouched
        """
        return func

    @staticmethod
    def ignored(func: Callable[Concatenate[T, P], U]) -> Callable[Concatenate[T, str, P], U]:
        """
        Call the function without textual arguments
        """
        return wrap_before_ignored(func)

    @staticmethod
    def quoted(mandatory: int,
               optional: int = 0,
               defaults: list[Any] | None = None,
               ignore_trailing_arguments: bool = False) -> Callable[[Callable[Concatenate[T, list[str], P], U]], Callable[Concatenate[T, str, P], U]]:
        """The function receives a list with a number of arguments that is between
        the numbers `mandatory` and `optional`.

        If the string doesn’t contain at least `mandatory` arguments, we return
        None because the given arguments are invalid.

        If there are any remaining arguments after `mandatory` and `optional`
        arguments have been found (and “ignore_trailing_arguments" is not True),
        we append them to the last argument of the list.

        An argument is a string (with or without whitespaces) between two quotes
        ("), or a whitespace separated word (if not inside quotes).

        The argument `defaults` is a list of strings that are used when an
        optional argument is missing.  For example if we accept one optional
        argument and none is provided, but we have one value in the `defaults`
        list, we use that string inplace. The `defaults` list can only
        replace missing optional arguments, not mandatory ones. And it
        should not contain more than `mandatory` values. Also you cannot

        Example:
        This method needs at least one argument, and accepts up to 3
        arguments

        >> @command_args_parser.quoted(1, 2, ['default for first arg'], False)
        >> def f(args):
        >>     print(args)

        >> f('coucou les amis') # We have one mandatory and two optional
        ['coucou', 'les', 'amis']
        >> f('"coucou les amis" "PROUT PROUT"') # One mandator and only one optional,
                                                # no default for the second
        ['coucou les amis', 'PROUT PROUT']
        >> f('')                # Not enough args for mandatory number
        None
        >> f('"coucou les potes"') # One mandatory, and use the default value
                                   # for the first optional
        ['coucou les potes, 'default for first arg']
        >> f('"un et demi" deux trois quatre cinq six') # We have three trailing arguments
        ['un et demi', 'deux', 'trois quatre cinq six']

        """
        default_args_outer = defaults or []

        def first(func: Callable[Concatenate[T, list[str], P], U]) -> Callable[Concatenate[T, str, P], U]:
            def before(params: str) -> list[str]:
                default_args = default_args_outer
                cmdargs = params
                if cmdargs and cmdargs.strip():
                    split_args = common.shell_split(cmdargs)
                else:
                    split_args = []
                if len(split_args) < mandatory:
                    return []
                res, split_args = split_args[:mandatory], split_args[
                    mandatory:]
                if optional == -1:
                    opt_args = split_args[:]
                else:
                    opt_args = split_args[:optional]

                if opt_args:
                    res += opt_args
                    split_args = split_args[len(opt_args):]
                    default_args = default_args[len(opt_args):]
                res += default_args
                if split_args and res and not ignore_trailing_arguments:
                    res[-1] += " " + " ".join(split_args)
                return res
            wrap = wrap_before_quoted(func, before=before)
            return wrap
        return first


command_args_parser = CommandArgParser()


def deny_anonymous(func: C) -> C:
    """Decorator to disable commands when using an anonymous account."""

    def before(args: list[Any], kwargs: dict[str, Any]) -> bool:
        core = args[0].core
        if core.xmpp.anon:
            core.information(
                'This command is not available for anonymous accounts.',
                'Info'
            )
            return False
        return True
    return cast(C, wrap_before_cancel_on_false(func, before=before))


def command_jid_from_context(func: Callable[[CoreSelf, JID | None], U]) -> Callable[[CoreSelf, list[str]], U | None]:
    """Decorator to get the ConversationsTab's JID, unless specified."""

    from slixmpp import JID, InvalidJID
    from poezio import tabs
    from poezio.contact import Contact, Resource

    def before(self: CoreSelf, args: list[str]) -> JID | None:
        if not hasattr(self, 'core'):
            return None
        jid = None
        if args and args[0]:
            try:
                jid = JID(args[0][0])
            except InvalidJID:
                self.core.information(f'Invalid JID {args[0][0]}', 'Error')
                return None

        current_tab = self.core.tabs.current_tab
        if jid is None:
            if isinstance(current_tab, tabs.RosterInfoTab):
                roster_tab = self.core.tabs.by_name_and_class(
                    'Roster',
                    tabs.RosterInfoTab,
                )
                if roster_tab is not None:
                    item = roster_tab.selected_row
                    if isinstance(item, Contact):
                        jid = item.bare_jid
                    elif isinstance(item, Resource):
                        jid = JID(item.jid)

            chattabs = (
                tabs.ConversationTab,
                tabs.StaticConversationTab,
                tabs.DynamicConversationTab,
            )
            if isinstance(current_tab, chattabs):
                jid = JID(current_tab.jid.bare)

        return jid

    def wrap(self: CoreSelf, args: list[str]) -> U | None:
        jid = before(self, args)
        if jid is None:
            self.core.information(f'No specified JID to {func.__name__}', 'Error')
            return None
        else:
            return func(self, jid)

    def awrap(self: CoreSelf, args: list[str]) -> U | None:
        jid = before(self, args)
        if jid is None:
            self.core.information(f'No specified JID to {func.__name__}', 'Error')
            return None
        else:
            return func(self, jid)

    if iscoroutinefunction(func):
        return cast(Callable[[CoreSelf, list[str]], U], awrap)
    else:
        return cast(Callable[[CoreSelf, list[str]], U], wrap)



"""
Bookmarks module

Therein the bookmark class is defined, representing one conference room.
This object is used to generate elements for both local and remote
bookmark storage. It can also parse xml Elements.

This module also defines several functions for retrieving and updating
bookmarks, both local and remote.

Poezio start scenario:

- upon initial connection, poezio will disco#info the server
- the available storage methods will be stored in the available_storage dict
    (either 'pep' or 'privatexml')
- if only one is available, poezio will set the use_bookmarks_method config option
    to it. If both are, it will be set to 'privatexml' (or if it was previously set, the
    value will be kept).
- it will then query the preferred storages for bookmarks and cache them locally
    (Bookmark objects with a method='remote' attribute)

Adding a remote bookmark:

- New Bookmark object added to the list with storage='remote'
- All bookmarks are sent to the storage selected in use_bookmarks_method
    if there was an error, the user is notified.


"""

import functools
import logging
from typing import (
    Any,
    Callable,
    List,
    Union,
)
from collections.abc import Iterator
from xml.etree import ElementTree as ET

from slixmpp import (
    InvalidJID,
    JID,
)
from slixmpp.exceptions import IqError, IqTimeout
from slixmpp import Iq
from slixmpp.plugins.xep_0048 import Bookmarks, Conference, URL
from poezio.connection import Connection
from poezio.config import config

log = logging.getLogger(__name__)


class Bookmark:
    def __init__(self,
                 jid: JID | str,
                 name: str | JID | None = None,
                 autojoin: bool = False,
                 nick: str | None = None,
                 password: str | None = None,
                 method: str = 'local') -> None:
        try:
            if isinstance(jid, JID):
                self._jid = jid
            else:
                self._jid = JID(jid)
        except InvalidJID:
            log.debug('Invalid JID %r provided for bookmark', jid)
            raise
        self.name = name or str(self.jid)
        self.autojoin = autojoin
        self.nick = nick
        self.password = password
        self._method = method

    @property
    def jid(self) -> JID:
        """Jid getter"""
        return self._jid

    @jid.setter
    def jid(self, jid: JID) -> None:
        try:
            if isinstance(jid, JID):
                self._jid = jid
            else:
                self._jid = JID(jid)
        except InvalidJID:
            log.debug('Invalid JID %r provided for bookmark', jid)
            raise

    @property
    def method(self) -> str:
        return self._method

    @method.setter
    def method(self, value: str) -> None:
        if value not in ('local', 'remote'):
            log.debug('Could not set bookmark storing method: %s', value)
            return
        self._method = value

    def __repr__(self) -> str:
        return '<%s%s|%s>' % (self.jid, ('/' + self.nick)
                              if self.nick else '', self.method)

    def stanza(self) -> Conference:
        """
        Generate a <conference/> stanza from the instance
        """
        el = Conference()
        if str(self.jid) != self.name:
            el['name'] = self.name
        el['jid'] = self.jid
        el['autojoin'] = 'true' if self.autojoin else 'false'
        if self.nick:
            el['nick'] = self.nick
        if self.password:
            el['password'] = self.password
        return el

    def local(self) -> str:
        """Generate a str for local storage"""
        local = str(self.jid)
        if self.nick:
            local += '/%s' % self.nick
        local += ':'
        if self.password:
            config.set_and_save('password', self.password, section=str(self.jid))
        return local

    @functools.singledispatchmethod
    @staticmethod
    def parse(el: ET.Element) -> 'Bookmark':
        """
        Generate a Bookmark object from a <conference/> element
        (this is a fallback for raw XML Elements)
        """
        jid = el.get('jid') or ''
        name = el.get('name')
        autojoin = True if el.get('autojoin',
                                  'false').lower() in ('true', '1') else False
        nick = None
        for n in el.iter('nick'):
            nick = n.text
        password = None
        for p in el.iter('password'):
            password = p.text

        return Bookmark(jid, name, autojoin, nick, password, method='remote')

    @staticmethod
    @parse.register(Conference)
    def parse_from_stanza(el: Conference) -> 'Bookmark':
        """
        Parse a Conference element into a Bookmark object
        """
        jid = el['jid']
        autojoin = el['autojoin']
        password = el['password']
        nick = el['nick']
        name = el['name']
        return Bookmark(jid, name, autojoin, nick, password, method='remote')


class BookmarkList:
    def __init__(self) -> None:
        self.bookmarks: list[Bookmark] = []
        preferred = config.getstr('use_bookmarks_method').lower()
        if preferred not in ('pep', 'privatexml'):
            preferred = 'privatexml'
        self.preferred = preferred
        self.available_storage = {
            'privatexml': False,
            'pep': False,
        }

    def __getitem__(self, key: str | JID | int) -> Bookmark | None:
        if isinstance(key, (str, JID)):
            for i in self.bookmarks:
                if key == i.jid:
                    return i
        elif isinstance(key, int):
            return self.bookmarks[key]
        return None

    def __contains__(self, key: str | JID | Bookmark) -> bool:
        if isinstance(key, (str, JID)):
            for bookmark in self.bookmarks:
                if bookmark.jid == key:
                    return True
        else:
            return key in self.bookmarks
        return False

    def remove(self, key: str | JID | Bookmark) -> None:
        if isinstance(key, (str, JID)):
            for i in self.bookmarks[:]:
                if i.jid == key:
                    self.bookmarks.remove(i)
        else:
            self.bookmarks.remove(key)

    def __iter__(self) -> Iterator[Bookmark]:
        return iter(self.bookmarks)

    def local(self) -> list[Bookmark]:
        return [bm for bm in self.bookmarks if bm.method == 'local']

    def remote(self) -> list[Bookmark]:
        return [bm for bm in self.bookmarks if bm.method == 'remote']

    def set(self, new: list[Bookmark]) -> None:
        self.bookmarks = new

    def append(self, bookmark: Bookmark) -> None:
        bookmark_exists = self[bookmark.jid]
        if not bookmark_exists:
            self.bookmarks.append(bookmark)
        else:
            self.bookmarks.remove(bookmark_exists)
            self.bookmarks.append(bookmark)

    def set_bookmarks_method(self, value: str) -> None:
        if self.available_storage.get(value):
            self.preferred = value
            config.set_and_save('use_bookmarks_method', value)

    async def save_remote(self, xmpp: Connection) -> Iq | None:
        """Save the remote bookmarks."""
        if not any(self.available_storage.values()):
            return None
        method = 'xep_0049' if self.preferred == 'privatexml' else 'xep_0223'

        if method:
            return await xmpp.plugin['xep_0048'].set_bookmarks(
                stanza_storage(self.bookmarks),
                method=method,
            )
        return None

    def save_local(self) -> None:
        """Save the local bookmarks."""
        local = ''.join(bookmark.local() for bookmark in self
                        if bookmark.method == 'local')
        config.set_and_save('rooms', local)

    async def save(self, xmpp: Connection, core=None) -> Iq | None:
        """Save all the bookmarks."""
        self.save_local()
        if config.getbool('use_remote_bookmarks'):
            try:
                result = await self.save_remote(xmpp)
            except (IqError, IqTimeout):
                if core is not None:
                    core.information(
                        'Could not save remote bookmarks.',
                        'Error'
                    )
                raise
            if core is not None:
                core.information('Bookmarks saved', 'Info')
            return result
        return None

    async def get_pep(self, xmpp: Connection) -> Iq:
        """Add the remotely stored bookmarks via pep to the list."""
        iq = await xmpp.plugin['xep_0048'].get_bookmarks(method='xep_0223')
        for conf in iq['pubsub']['items']['item']['bookmarks'][
                'conferences']:
            if isinstance(conf, URL):
                continue
            bookm = Bookmark.parse(conf)
            self.append(bookm)
        return iq

    async def get_privatexml(self, xmpp: Connection) -> Iq:
        """
        Fetch the remote bookmarks stored via privatexml.
        """

        iq = await xmpp.plugin['xep_0048'].get_bookmarks(method='xep_0049')
        for conf in iq['private']['bookmarks']['conferences']:
            bookm = Bookmark.parse(conf)
            self.append(bookm)
        return iq

    async def get_remote(self, xmpp: Connection, information: Callable[[str, str], Any]) -> Iq | None:
        """Add the remotely stored bookmarks to the list."""
        if xmpp.anon or not any(self.available_storage.values()):
            information('No remote bookmark storage available', 'Warning')
            return None
        if self.preferred == 'pep':
            return await self.get_pep(xmpp)
        else:
            return await self.get_privatexml(xmpp)

    def get_local(self) -> None:
        """Add the locally stored bookmarks to the list."""
        rooms = config.getlist('rooms')
        if not rooms:
            return
        for room in rooms:
            try:
                jid = JID(room)
            except InvalidJID:
                continue
            if jid.bare == '':
                continue
            if jid.resource != '':
                nick = jid.resource
            else:
                nick = None
            passwd = config.get_by_tabname(
                'password', JID(jid.bare), fallback=False) or None
            b = Bookmark(
                jid.bare,
                jid.user,
                autojoin=True,
                nick=nick,
                password=passwd,
                method='local')
            self.append(b)


def stanza_storage(bookmarks: Union[BookmarkList, List[Bookmark]]) -> Bookmarks:
    """Generate a <storage/> stanza with the conference elements."""
    storage = Bookmarks()
    for b in (b for b in bookmarks if b.method == 'remote'):
        storage.append(b.stanza())
    return storage



"""
This plugin allows sending message retractions on our own messages.

Command
-------

.. glossary::

    /retract
        **Usage:** ``/retract <message time>``
"""

from poezio.decorators import command_args_parser
from poezio.plugin import BasePlugin
from poezio.core.structs import Completion
from poezio.windows.inputs import Input
from poezio.ui.types import Message
from poezio import tabs


DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S'


class Plugin(BasePlugin):
    """Message Retraction (XEP-0424) plugin"""

    def init(self) -> None:
        self.core.xmpp.register_plugin('xep_0424')

        for _tabtype in (tabs.MucTab, tabs.DynamicConversationTab,
                         tabs.StaticConversationTab, tabs.PrivateTab):
            self.api.add_tab_command(
                _tabtype,
                name='retract',
                handler=self.command_retract,
                usage='<message time>',
                help='Retract one of your messages using its timestamp (using tab complete)',
                short='Retract a message using its timestamp',
                completion=self.completion_retract,
            )

    def completion_retract(self, the_input: Input) -> Completion | None:
        """Datetime completion"""
        if the_input.get_argument_position() != 1:
            return None
        tab = self.api.current_tab()
        if not isinstance(tab, tabs.ChatTab):
            return None

        all_messages = self.api.get_conversation_messages()[:-100:-1]
        # We want our own messages only
        if isinstance(tab, tabs.MucTab):
            filtered_messages = [
                m for m in all_messages
                if isinstance(m, Message) and m.user and m.user == tab.own_user
            ]
        else:
            filtered_messages = [
                m for m in all_messages
                if isinstance(m, Message) and m.jid and m.jid.bare == self.core.xmpp.boundjid.bare
            ]
        messages = list(map(lambda m: m.time.strftime(DATETIME_FORMAT), filtered_messages))

        if not messages:
            return None

        return Completion(the_input.auto_completion, messages, quotify=False)

    @command_args_parser.raw
    async def command_retract(self, args: str) -> None:
        """Moderate a message in a chatroom"""
        if not args:
            return
        time = args.strip()

        tab = self.api.current_tab()
        if not isinstance(tab, tabs.ChatTab):
            return
        is_muc = isinstance(tab, tabs.MucTab)

        found_msg = self.api.find_message(time, DATETIME_FORMAT)
        # TODO: do better here
        # cannot re-use is_muc because mypy is not smart enough
        if isinstance(tab, tabs.MucTab) and found_msg and found_msg.user != tab.own_user:
            self.api.information(
                'Unable to retract message: duplicate timestamps',
                'Error',
            )
        if not found_msg:
            self.api.information(
                f'Message with timestamp “{time}” not found.',
                'Error',
            )
            return

        mtype = 'chat'
        id_ = found_msg.identifier
        if is_muc:
            id_ = found_msg.stanza_id or ''
            mtype = 'groupchat'
            if not id_:
                self.api.information(
                    f'Message with timestamp “{time}” '
                    'found but contained no “{urn:xmpp:sid:0}stanza-id” identifier.',
                    'Error',
                )
                return

        jid = tab.jid
        self.core.xmpp['xep_0424'].send_retraction(jid, id_, mtype=mtype)
        if isinstance(tab, tabs.OneToOneTab) and found_msg.jid:
            tab.handle_retract_values(id_, found_msg.jid)



"""
Show the exchanged IQs (useful for debugging).

"""
from poezio.plugin import BasePlugin
from slixmpp.xmlstream import StanzaBase
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.xmlstream.handler import Callback


class Plugin(BasePlugin):
    def init(self) -> None:
        self.core.xmpp.register_handler(
            Callback('Iq_show', StanzaPath('iq'), self.handle_iq))

    def handle_iq(self, iq: StanzaBase) -> None:
        self.api.information('%s' % iq, 'Iq')

    def cleanup(self) -> None:
        self.core.xmpp.remove_handler('Iq_show')



# Copyright 2010-2011 Florent Le Coz <louiz@louiz.org>
#
# This file is part of Poezio.
#
# Poezio is free software: you can redistribute it and/or modify
# it under the terms of the GPL-3.0+ license. See the COPYING file.
"""
Various useful functions.
"""

from datetime import (
    datetime,
    timedelta,
    timezone,
)
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union
from dataclasses import dataclass

import os
import subprocess
import time
import string
import logging
import itertools

from slixmpp import Iq, Message, Presence, JID
from poezio.poezio_shlex import shlex

log = logging.getLogger(__name__)


def _get_output_of_command(command: str) -> Optional[List[str]]:
    """
    Runs a command and returns its output.

    :param str command: The command to run.
    :return: The output or None
    :rtype: :py:class:`str`
    """
    try:
        return subprocess.check_output(
            command.split()).decode('utf-8').split('\n')
    except subprocess.CalledProcessError:
        return None


def _is_in_path(command: str, return_abs_path: bool = False) -> Union[bool, str]:
    """
    Check if *command* is in the $PATH or not.

    :param str command: The command to be checked.
    :param bool return_abs_path: Return the absolute path of the command instead
        of True if the command is found.
    :return: True if the command is found, the command path if the command is found
        and *return_abs_path* is True, otherwise False.

    """
    path = os.getenv('PATH', default='')
    for directory in path.split(os.pathsep):
        try:
            if command in os.listdir(directory):
                if return_abs_path:
                    return os.path.join(directory, command)
                return True
        except OSError:
            # If the user has non directories in his path
            pass
    return False


DISTRO_INFO = {
    'Arch Linux': Path('/etc/arch-release'),
    'Aurox Linux': Path('/etc/aurox-release'),
    'Conectiva Linux': Path('/etc/conectiva-release'),
    'CRUX': Path('/usr/bin/crux'),
    'Debian GNU/Linux': Path('/etc/debian_version'),
    'Fedora Linux': Path('/etc/fedora-release'),
    'Gentoo Linux': Path('/etc/gentoo-release'),
    'Linux from Scratch': Path('/etc/lfs-release'),
    'Mandrake Linux': Path('/etc/mandrake-release'),
    'Slackware Linux': Path('/etc/slackware-version'),
    'Solaris/Sparc': Path('/etc/release'),
    'Source Mage': Path('/etc/sourcemage_version'),
    'SUSE Linux': Path('/etc/SuSE-release'),
    'Sun JDS': Path('/etc/sun-release'),
    'PLD Linux': Path('/etc/pld-release'),
    'Yellow Dog Linux': Path('/etc/yellowdog-release'),
    # many distros use the /etc/redhat-release for compatibility
    # so Redhat is the last
    'Redhat Linux': Path('/etc/redhat-release')
}


def get_os_info() -> str:
    """
    Returns a detailed and well formatted string containing
    information about the operating system

    :rtype: str
    """
    if os.name == 'posix':
        executable = 'lsb_release'
        params = ' --description --codename --release --short'
        full_path_to_executable = _is_in_path(executable, return_abs_path=True)
        if full_path_to_executable:
            command = executable + params
            process = subprocess.Popen(
                [command],
                shell=True,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                close_fds=True)
            process.wait()
            if process.stdout is not None:
                out = process.stdout.readline().decode('utf-8').strip()
                # some distros put n/a in places, so remove those
                out = out.replace('n/a', '').replace('N/A', '')
                return out
            return ''

        # lsb_release executable not available, so parse files
        for distro_name in DISTRO_INFO:
            path_to_file = DISTRO_INFO[distro_name]
            if path_to_file.exists():
                if os.access(str(path_to_file), os.X_OK):
                    # the file is executable (f.e. CRUX)
                    # yes, then run it and get the first line of output.
                    output = _get_output_of_command(str(path_to_file))
                    if output:
                        text = output[0]
                    else:
                        text = ''
                else:
                    with path_to_file.open(encoding='utf-8') as fdes:
                        text = fdes.readline().strip()  # get only first line
                    basename = path_to_file.name
                    if basename.endswith('version'):
                        # sourcemage_version and slackware-version files
                        # have all the info we need (name and version of distro)
                        if not basename.startswith('sourcemage') or not\
                                basename.startswith('slackware'):
                            text = distro_name + ' ' + text
                    elif basename == 'aurox-release' or \
                            basename == 'arch-release':
                        # file doesn't have version
                        text = distro_name
                    elif basename == 'lfs-release':
                        # file just has version
                        text = distro_name + ' ' + text
                os_info = text.replace('\n', '')
                return os_info

        # our last chance, ask uname and strip it
        uname_output = _get_output_of_command('uname -sr')
        if uname_output is not None:
            os_info = uname_output[0]  # only first line
            return os_info
    os_info = 'N/A'
    return os_info


def _datetime_tuple(timestamp: str) -> datetime:
    """
    Convert a timestamp using strptime and the format: %Y%m%dT%H:%M:%S.

    Because various datetime formats are used, the following exceptions
    are handled:

    * Optional milliseconds appended to the string are removed
    * Optional Z (that means UTC) appended to the string are removed
    * XEP-082 datetime strings have all '-' chars removed to meet the above format.

    :param str timestamp: The string containing the formatted date.
    :return: The date.
    :rtype: :py:class:`datetime.datetime`
    """
    timestamp = timestamp.replace('-', '', 2).replace(':', '')
    date = timestamp[:15]
    tz_msg = timestamp[15:]
    try:
        ret = datetime.strptime(date, '%Y%m%dT%H%M%S')
    except ValueError:
        ret = datetime.now()
    # add the message timezone if any
    if tz_msg and tz_msg != 'Z':
        tz_mod = -1 if tz_msg[0] == '-' else 1
        try:
            tz_parsed = time.strptime(tz_msg[1:], '%H%M')
        except ValueError:
            pass  # ignore if we got a badly-formatted offset
        else:
            tz_seconds = tz_parsed.tm_hour * 3600 + tz_parsed.tm_min * 60
            delta = timedelta(seconds=tz_mod * tz_seconds)
            ret -= delta
    # convert UTC to local time, with DST etc.
    if time.daylight and time.localtime().tm_isdst:
        tz = timedelta(seconds=-time.altzone)
    else:
        tz = timedelta(seconds=-time.timezone)
    ret += tz
    return ret


def get_utc_time(local_time: Optional[datetime] = None) -> datetime:
    """
    Get the current UTC time

    :param datetime local_time: The current local time
    :return: The current UTC time
    """
    if local_time is None:
        local_time = datetime.now()
        isdst = time.localtime().tm_isdst
    else:
        isdst = time.localtime(int(local_time.timestamp())).tm_isdst

    if time.daylight and isdst:
        tz = timedelta(seconds=time.altzone)
    else:
        tz = timedelta(seconds=time.timezone)

    utc_time = local_time + tz

    return utc_time


def get_local_time(utc_time: datetime) -> datetime:
    """
    Get the local time from an UTC time
    """
    isdst = time.localtime(int(utc_time.timestamp())).tm_isdst

    if time.daylight and isdst:
        tz = timedelta(seconds=time.altzone)
    else:
        tz = timedelta(seconds=time.timezone)

    local_time = utc_time - tz

    return local_time


def find_delayed_tag(message: Message) -> Tuple[bool, Optional[datetime]]:
    """
    Check if a message is delayed or not.

    :param slixmpp.Message message: The message to check.
    :return: A tuple containing (True, the datetime) or (False, None)
    :rtype: :py:class:`tuple`
    """

    find_delay = message.xml.find
    delay_tag = message.get_plugin('delay', check=True)
    date: Optional[datetime] = None
    if delay_tag is not None:
        delayed = True
        date = delay_tag['stamp']
        # Convert a tz-aware date to localtime
        if date.tzinfo:
            # convert UTC to local time, with DST etc.
            if time.daylight and time.localtime().tm_isdst:
                tz = timedelta(seconds=-time.altzone)
            else:
                tz = timedelta(seconds=-time.timezone)
            offset = date.tzinfo.utcoffset(date) or timedelta(0)
            date = date.replace(tzinfo=None) + offset + tz
    else:
        # We support the OLD and deprecated XEP: http://xmpp.org/extensions/xep-0091.html
        # But it sucks, please, Jabber servers, don't do this :(
        raw_delay_tag = find_delay('{jabber:x:delay}x')
        if raw_delay_tag is not None:
            delayed = True
            date = _datetime_tuple(raw_delay_tag.attrib['stamp'])
        else:
            delayed = False
    return (delayed, date)


def shell_split(st: str) -> List[str]:
    """
    Split a string correctly according to the quotes
    around the elements.

    :param str st: The string to split.
    :return: A list of the different of the string.
    :rtype: :py:class:`list`

    >>> shell_split('"sdf 1" "toto 2"')
    ['sdf 1', 'toto 2']
    """
    sh = shlex(st)
    ret = []
    w = sh.get_token()
    while w and w[2] is not None:
        ret.append(w[2])
        if w[1] == len(st):
            return ret
        w = sh.get_token()
    return ret


def find_argument(pos: int, text: str, quoted: bool = True) -> int:
    """
    Split an input into a list of arguments, return the number of the
    argument selected by pos.

    If the position searched is outside the string, or in a space between words,
    then it will return the position of an hypothetical new argument.

    See the doctests of the two methods for example behaviors.

    :param int pos: The position to search.
    :param str text: The text to analyze.
    :param bool quoted: Whether to take quotes into account or not.
    :rtype: int
    """
    if quoted:
        return _find_argument_quoted(pos, text)
    return _find_argument_unquoted(pos, text)


def _find_argument_quoted(pos: int, text: str) -> int:
    """
    Get the number of the argument at position pos in
    a string with possibly quoted text.
    """
    sh = shlex(text)
    count = -1
    w = sh.get_token()
    while w and w[2] is not None:
        count += 1
        if w[0] <= pos < w[1]:
            return count
        w = sh.get_token()

    return count + 1


def _find_argument_unquoted(pos: int, text: str) -> int:
    """
    Get the number of the argument at position pos in
    a string without interpreting quotes.
    """
    ret = text.split()
    search = 0
    argnum = 0
    for i, elem in enumerate(ret):
        elem_start = text.find(elem, search)
        elem_end = elem_start + len(elem)
        search = elem_end
        if elem_start <= pos < elem_end:
            return i
        argnum = i
    return argnum + 1


def parse_str_to_secs(duration: str = '') -> int:
    """
    Parse a string of with a number of d, h, m, s.

    :param str duration: The formatted string.
    :return: The number of seconds represented by the string
    :rtype: :py:class:`int`

    >>> parse_str_to_secs("1d3m1h")
    90180
    """
    values = {'s': 1, 'm': 60, 'h': 3600, 'd': 86400}
    result = 0
    tmp = '0'
    for char in duration:
        if char in string.digits:
            tmp += char
        elif char in values:
            tmp_i = int(tmp)
            result += tmp_i * values[char]
            tmp = '0'
        else:
            return 0
    if tmp != '0':
        result += int(tmp)
    return result


def parse_secs_to_str(duration: int = 0) -> str:
    """
    Do the reverse operation of :py:func:`parse_str_to_secs`.

    Parse a number of seconds to a human-readable string.
    The string has the form XdXhXmXs. 0 units are removed.

    :param int duration: The duration, in seconds.
    :return: A formatted string containing the duration.
    :rtype: :py:class:`str`

    >>> parse_secs_to_str(3601)
    '1h1s'
    """
    secs, mins, hours, days = 0, 0, 0, 0
    result = ''
    secs = duration % 60
    mins = (duration % 3600) // 60
    hours = (duration % 86400) // 3600
    days = duration // 86400

    result += '%dd' % days if days else ''
    result += '%dh' % hours if hours else ''
    result += '%dm' % mins if mins else ''
    result += '%ds' % secs if secs else ''
    if not result:
        result = '0s'
    return result


def format_tune_string(infos: Dict[str, str]) -> str:
    """
    Construct a string from a dict created from an "User tune" event.

    :param dict infos: Tune information
    :return: The formatted string
    :rtype: :py:class:`str`
    """
    elems = []
    track = infos.get('track')
    if track:
        elems.append(track)
    title = infos.get('title')
    if title:
        elems.append(title)
    else:
        elems.append('Unknown title')
    elems.append('-')
    artist = infos.get('artist')
    if artist:
        elems.append(artist)
    else:
        elems.append('Unknown artist')

    rating = infos.get('rating')
    if rating:
        elems.append('[ ' + rating + '/10 ]')
    length_str = infos.get('length')
    if length_str:
        length = int(length_str)
        secs = length % 60
        mins = length // 60
        secs_str = str(secs).zfill(2)
        mins_str = str(mins).zfill(2)
        elems.append('[' + mins_str + ':' + secs_str + ']')
    return ' '.join(elems)


def format_gaming_string(infos: Dict[str, str]) -> str:
    """
    Construct a string from a dict containing "user gaming" information.
    (for now, only use address and name)

    :param dict infos: Gaming information
    :returns: The formatted string
    :rtype: :py:class:`str`
    """
    name = infos.get('name')
    if not name:
        return ''

    server_address = infos.get('server_address')
    if server_address:
        return '%s on %s' % (name, server_address)
    return name


def unique_prefix_of(a: str, b: str) -> str:
    """
    Return the unique prefix of `a` with `b`.

    Corner cases:

    * If `a` and `b` share no prefix, the first letter of `a` is returned.
    * If `a` and `b` are equal, `a` is returned.
    * If `a` is a prefix of `b`, `a` is returned.
    * If `b` is a prefix of `a`, `b` plus the first letter of `a` after the
      common prefix is returned.
    """
    for i, (ca, cb) in enumerate(itertools.zip_longest(a, b)):
        if ca != cb:
            return a[:i+1]
    # both are equal, return a
    return a


def to_utc(time_: datetime) -> datetime:
    """Convert a datetime-aware time zone into raw UTC"""
    if time_.tzinfo is not None:  # Convert to UTC
        time_ = time_.astimezone(tz=timezone.utc)
    else:  # Assume local tz, convert to UTC
        tzone = datetime.now().astimezone().tzinfo
        time_ = time_.replace(tzinfo=tzone).astimezone(tz=timezone.utc)
    # Return an offset-naive datetime
    return time_.replace(tzinfo=None)


# http://xmpp.org/extensions/xep-0045.html#errorstatus
ERROR_AND_STATUS_CODES = {
    '401': 'A password is required',
    '403': 'Permission denied',
    '404': 'The room doesn’t exist',
    '405': 'Your are not allowed to create a new room',
    '406': 'A reserved nick must be used',
    '407': 'You are not in the member list',
    '409': 'This nickname is already in use or has been reserved',
    '503': 'The maximum number of users has been reached',
}


# http://xmpp.org/extensions/xep-0086.html
DEPRECATED_ERRORS = {
    '302': 'Redirect',
    '400': 'Bad request',
    '401': 'Not authorized',
    '402': 'Payment required',
    '403': 'Forbidden',
    '404': 'Not found',
    '405': 'Not allowed',
    '406': 'Not acceptable',
    '407': 'Registration required',
    '408': 'Request timeout',
    '409': 'Conflict',
    '500': 'Internal server error',
    '501': 'Feature not implemented',
    '502': 'Remote server error',
    '503': 'Service unavailable',
    '504': 'Remote server timeout',
    '510': 'Disconnected',
}


def get_error_message(stanza: Message | Iq | Presence, deprecated: bool = False) -> str:
    """
    Takes a stanza of the form <message type='error'><error/></message>
    and return a well formed string containing error information
    """
    sender = stanza['from']
    msg = stanza['error']['type']
    condition = stanza['error']['condition']
    code = stanza['error']['code']
    body = stanza['error']['text']
    if not body:
        if deprecated:
            if code in DEPRECATED_ERRORS:
                body = DEPRECATED_ERRORS[code]
            else:
                body = condition or 'Unknown error'
        else:
            if code in ERROR_AND_STATUS_CODES:
                body = ERROR_AND_STATUS_CODES[code]
            else:
                body = condition or 'Unknown error'
    if code:
        message = '%(from)s: %(code)s - %(msg)s: %(body)s' % {
            'from': sender,
            'msg': msg,
            'body': body,
            'code': code
        }
    else:
        message = '%(from)s: %(msg)s: %(body)s' % {
            'from': sender,
            'msg': msg,
            'body': body
        }
    return message


class PresenceError(Exception):
    pass


@dataclass(slots=True)
class DissectedPresence:
    from_nick: str
    from_room: str
    affiliation: str
    show: str
    status: str
    role: str
    jid: JID | None
    typ: str
    occupant_id: str


def dissect_presence(presence: Presence, include_occupantid: bool = False) -> DissectedPresence:
    """
    Extract relevant information from a presence
    """
    from_nick = presence['from'].resource
    from_room = presence['from'].bare
    # Check if it's not an error presence.
    if presence['type'] == 'error':
        raise PresenceError(presence)
    affiliation = presence['muc']['affiliation']
    show = presence['show']
    status = presence['status']
    role = presence['muc']['role']
    jid = presence['muc']['jid']
    typ = presence['type']
    occupant_id = ''
    if include_occupantid:
        occupant_id_element = presence['occupant-id']
        if occupant_id_element:
            occupant_id = occupant_id_element['id']
    return DissectedPresence(
        from_nick, from_room, affiliation, show, status, role,
        jid, typ, occupant_id,
    )



"""
This plugin is here to send what you are listening to in a chat tab.

Installation
------------

You need `python-mpd`_, in its python3 version.

Then you can load the plugin.

.. code-block:: none

 /load mpd_client


Configuration
-------------

You have to put the following into :file:`mpd_client.cfg`, as explained in
the :ref:`plugin-configuration` section.

.. note:: If you do not put anything, the plugin will try to connect to
        :file:`localhost:6600` with no password.

.. code-block:: ini

    [mpd_client]
    host = the_mpd_host
    port = 6600
    password = password if necessary


Usage
-----

.. glossary::

    /mpd
        **Usage:** ``/mpd [full]``

        The bare command will show the current song, artist, and album

        ``/mpd full`` will show the current song, artist, and album,
        plus a nice progress bar in color.

.. _python-mpd: https://github.com/Mic92/python-mpd2

"""

from poezio.plugin import BasePlugin
from poezio.common import shell_split
from poezio.core.structs import Completion
from poezio.windows.inputs import Input
from os.path import basename as base
from poezio import tabs
try:
    import musicpd as mpd
except ImportError:
    mpd = None


class MPDLibNotFound(ImportError):
    def __init__(self) -> None:
        super().__init__("musicpd library not found")


class Plugin(BasePlugin):
    def init(self) -> None:
        if mpd is None:
            raise MPDLibNotFound()
        for _class in (tabs.DynamicConversationTab, tabs.StaticConversationTab, tabs.MucTab, tabs.PrivateTab):
            self.api.add_tab_command(
                _class,
                'mpd',
                self.command_mpd,
                usage='[full]',
                help=(
                    'Sends a message showing the current song of an MPD '
                    'instance. If full is provided, the message is more '
                    'verbose.'
                ),
                short='Send the MPD status',
                completion=self.completion_mpd)

    def command_mpd(self, str_args: str) -> None:
        args = shell_split(str_args)
        c = mpd.MPDClient()
        c.connect(
            host=self.config.get('host', 'localhost'),
            port=self.config.get('port', '6600'))
        password = self.config.get('password', '')
        if password:
            c.password(password)  # pylint: disable=no-member
        current = c.currentsong()  # pylint: disable=no-member
        artist = current.get('artist', 'Unknown artist')
        album = current.get('album', 'Unknown album')
        title = current.get('title', base(
            current.get('file', 'Unknown title')))

        s = '%s - %s (%s)' % (artist, title, album)
        if 'full' in args:
            if 'elapsed' in current and 'time' in current:
                current_time = float(c.status()['elapsed'])  # pylint: disable=no-member
                percents = int(current_time / float(current['time']) * 10)
                s += ' \x192}[\x191}' + '-' * (
                    percents - 1) + '\x193}+' + '\x191}' + '-' * (
                        10 - percents - 1) + '\x192}]\x19o'
        if not self.api.send_message('%s' % (s, )):
            self.api.information('Cannot send result (%s)' % s, 'Error')

    def completion_mpd(self, the_input: Input) -> Completion:
        return Completion(the_input.auto_completion, ['full'], quotify=False)



"""
This plugin colors each character of a message in white.

Usage
-----

.. glossary::

    /load white

        Say something in a Chat tab.

.. note:: This plugin is best used when someone else is writing in black,
assuming everyone is using a white background.  Black backgrounds matter too!
"""

from poezio.plugin import BasePlugin
from poezio import xhtml
from slixmpp import Message
from typing import Any


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_event_handler('muc_say', self.whiteify)
        self.api.add_event_handler('private_say', self.whiteify)
        self.api.add_event_handler('conversation_say', self.whiteify)

    @staticmethod
    def whiteify(msg: Message, _: Any) -> None:
        msg['body'] = '\x197}' + xhtml.clean_text(msg['body'])



"""
This plugin will set the title of your terminal to the name of the current tab.

"""
from poezio.plugin import BasePlugin
from poezio import tabs
import sys


class Plugin(BasePlugin):
    def init(self) -> None:
        self.on_tab_change(None, new_tab=self.core.tabs.current_tab)
        self.api.add_event_handler('tab_change', self.on_tab_change)

    def cleanup(self) -> None:
        "Re-set the terminal title to 'poezio'"
        sys.stdout.write("\x1b]0;poezio\x07")
        sys.stdout.flush()

    def on_tab_change(self, old_tab: tabs.Tab | None, new_tab: tabs.Tab) -> None:
        sys.stdout.write("\x1b]0;{}\x07".format(new_tab.name))
        sys.stdout.flush()



"""
Usage
-----

This plugin defines two new global commands: :term:`/alias` and :term:`/unalias`.

.. glossary::

    /alias
        **Usage:** ``/alias <name> <command> [args]``

        This command will create a new command, named ``<name>`` (and callable
        with ``/name``), that runs ``/command``, with ``[args]`` as fixed
        args for the command.
        When you run the alias, you can also pass parameters to it, that will be
        given to the original command.

        Example: ::

            /alias toto say koin

        Will bind ``/say koin`` to ``/toto``, so this alias will work in any
        Chat tab. If someone calls it with ::

            /toto koin

        Poezio will then execute ``/say koin koin``.

        Also, you can rebind arguments arbitrarily, with the ``{}`` placeholder.
        For example, ::

            /alias toto say {} le {}
            /toto loulou coucou

        Will execute ``/say loulou le coucou``, because the ``{}`` are
        replaced with the command args, in the order they are given.

        Extra args are still added at the end of the command if provided
        (args used for the formatting are only used for the formatting).

    /unalias
        **Usage:** ``/unalias <name>``

        This command removes a defined alias.


Config
------

The aliases are stored inside the configuration file for the plugin.
You can either use the above commands or write it manually, and it
will be read when the plugin is loaded.


Example of the syntax:

.. code-block:: ini

    [alias]
    toto = say {} le {}
    j = join {}@conference.jabber.org/nick
    jp = say je proteste


"""

from typing import Callable, Any, ParamSpec, TypeVar, Concatenate

from poezio.plugin import BasePlugin
from poezio.common import shell_split
from poezio.core.structs import Completion
from poezio.windows.inputs import Input


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_command(
            'alias',
            self.command_alias,
            usage='<alias> <command> [args]',
            short='Create an alias command',
            help='Create an alias for <command> with [args].')
        self.api.add_command(
            'unalias',
            self.command_unalias,
            usage='<alias>',
            help='Remove a previously created alias',
            short='Remove an alias',
            completion=self.completion_unalias)
        self.commands: dict[str, Callable[..., Any]] = {}
        self.load_conf()

    def load_conf(self) -> None:
        """
        load stored aliases on startup
        """
        for alias in self.config.options():
            full = self.config.get(alias, '')
            if full:
                self.command_alias(alias + ' ' + full, silent=True)

    def command_alias(self, line: str, silent: bool = False) -> None:
        """
        /alias <alias> <command> [args]
        """
        arg = split_args(line)
        if not arg:
            if not silent:
                self.api.information('Alias: Not enough parameters', 'Error')
            return
        alias, command, args = arg

        if alias in self.commands:
            update = True
        elif alias in self.core.commands:
            if not silent:
                self.api.information('Alias: command already exists', 'Error')
            return
        else:
            update = False

        self.config.set(alias, command + ' ' + args)
        self.commands[alias] = command_wrapper(
            generic_command,
            lambda: self.get_command(command),
            args,
        )
        self.api.del_command(alias)
        self.api.add_command(
            alias, self.commands[alias],
            'This command is an alias for /%s %s' % (alias, command))

        if not silent:
            if update:
                self.api.information('Alias /%s updated' % alias, 'Info')
            else:
                self.api.information('Alias /%s successfully created' % alias,
                                     'Info')

    def command_unalias(self, alias: str) -> None:
        """
        /unalias <existing alias>
        """
        if alias in self.commands:
            del self.commands[alias]
            self.api.del_command(alias)
            self.config.remove(alias)
            self.api.information('Alias /%s successfully deleted' % alias,
                                 'Info')

    def completion_unalias(self, the_input: 'Input') -> Completion:
        "Completion for /unalias"
        aliases = [alias for alias in self.commands]
        aliases.sort()
        return Completion(
            the_input.auto_completion, aliases, '', quotify=False)

    def get_command(self, name: str) -> Callable[[str], None]:
        """Returns the function associated with a command"""

        def dummy(args: str) -> None:
            """Dummy function called if the command doesn’t exist"""
            pass

        if name in self.commands:
            return dummy
        elif name in self.core.commands:
            return self.core.commands[name].func
        elif name in self.api.current_tab().commands:
            return self.api.current_tab().commands[name].func
        return dummy


def split_args(line: str) -> tuple[str, str, str] | None:
    """
    Extract the relevant vars from the command line
    """
    arg = line.split()
    if len(arg) < 2:
        return None
    alias_pos = line.find(' ')
    alias = line[:alias_pos]
    end = line[alias_pos + 1:]
    args_pos = end.find(' ')
    if args_pos == -1:
        command = end
        args = ''
    else:
        command = end[:args_pos]
        args = end[args_pos + 1:]
    return (alias, command, args)


def generic_command(command: Callable[[], Callable[[str], None]],
                    extra_args: str, args: str, /) -> None:
    """
    Function that will execute the command and set the relevant
    parameters (format string, etc).
    """
    list_args = shell_split(args)
    new_extra_args = extra_args.format(*list_args)
    count = extra_args.count('{}')
    list_args = list_args[count:]
    new_extra_args += ' '.join(list_args)
    return command()(new_extra_args)


P = ParamSpec('P')
T = TypeVar('T')
U = TypeVar('U')


def command_wrapper(
        func: Callable[[Callable[[], Callable[Concatenate[str, P], T]], str, str], T],
        command: Callable[[], Callable[Concatenate[str, P], T]],
        extra_args: str
) -> Callable[Concatenate[str, P], T]:
    "set the predefined arguments"

    def wrapper(arg: str, /, *args: P.args, **kwargs: P.kwargs) -> T:
        return func(command, extra_args, arg, **kwargs)

    return wrapper



"""
This plugin adds a :term:`/code` command, to send syntax highlighted snippets
of code using pygments and XHTML-IM (XEP-0071).

Install
-------

Either use your distribution tools to install python3-pygments or equivalent,
or run:

.. code-block:: shell

    pip install --user pygments

Usage
-----

.. glossary::

    /code <language> <snippet>

        Run this command to send the <snippet> of code, syntax highlighted
        using pygments’s <language> lexer.
"""

from poezio.plugin import BasePlugin
from poezio import tabs

from pygments import highlight
from pygments.lexers import get_lexer_by_name
from pygments.formatters import HtmlFormatter  # pylint: disable=no-name-in-module
FORMATTER = HtmlFormatter(nowrap=True, noclasses=True)


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_command(
            'code',
            self.command_code,
            usage='<language> <code>',
            short='Sends syntax-highlighted code',
            help='Sends syntax-highlighted code in the current tab')

    def command_code(self, args: str) -> None:
        split = args.split(None, 1)
        if len(split) != 2:
            self.api.information('Usage: /code <language> <code>', 'Error')
            return None
        language, code = split
        lexer = get_lexer_by_name(language)
        tab = self.api.current_tab()
        code = highlight(code, lexer, FORMATTER)
        if isinstance(tab, tabs.ChatTab):
            tab.command_xhtml(
                '<pre><code class="language-%s">%s</code></pre>' % (language, code.rstrip('\n')))



"""
Plugin destined to be used together with the Biboumi IRC gateway.

For more information about Biboumi, please see the `official website`_.

This plugin is here as a non-default extension of the poezio configuration
made to work with IRC rooms and logins. It also defines commands aimed at
reducing the amount of effort needed to navigate smoothly between IRC and
XMPP rooms.

Configuration
-------------

Every feature of this plugin is centered around its :ref:`configuration file <plugin-configuration>`,
so you have to make sure it is filled properly.

Global configuration
~~~~~~~~~~~~~~~~~~~~
.. glossary::
    :sorted:

    gateway
        **Default:** ``irc.jabberfr.org``

        The JID of the IRC gateway to use. If empty, irc.jabberfr.org will be
        used. Please try to run your own, though, it’s painless to setup.

    initial_connect
        **Default:** ``true``

        Set to ``true`` if you want to join all the rooms and try to
        authenticate with nickserv when the plugin gets loaded. If it set to
        ``false``, you will have to use the :term:`/irc_login` command to
        authenticate, and the :term:`/irc_join` command to join
        preconfigured rooms.

.. note:: There is no nickname option because the default from poezio will be used.

Server-specific configuration
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Write a configuration section for each server, with the server address as the
section name, and the following options:


.. glossary::
    :sorted:

    nickname
        **Default:** ``[empty]``

        Your nickname on this server. If empty, the default configuration will be used.

    rooms [IRC plugin]
        **Default:** ``[empty]``

        The list of rooms to join on this server (e.g. ``#room1:#room2``).

.. note:: If no login_command or login_nick is set, the authentication phase
        won’t take place and you will join the rooms without authentication
        with nickserv or whatever.

Commands
~~~~~~~~

.. glossary::
    :sorted:

    /irc_join
        **Usage:** ``/irc_join <room or server>``

        Join the specified room on the same server as the current tab (can
        be a private conversation or a chatroom). If a server that appears
        in the conversation is specified instead of a room, the plugin
        will try to join all the rooms configured with autojoin on that
        server.

    /irc_query
        **Usage:** ``/irc_query <nickname> [message]``

        Open a private conversation with the given nickname, on the same IRC
        server as the current tab (can be a private conversation or a
        chatroom).  Doing `/irc_query foo "hello there"` when the current
        tab is #foo%irc.example.com@biboumi.example.com is equivalent to
        ``/message foo%irc.example.com@biboumi.example.com "hello there"``

Example configuration
~~~~~~~~~~~~~~~~~~~~~

.. code-block:: ini

    [irc]
    gateway = irc.jabberfr.org

    [irc.libera.chat]
    nickname = mynick
    login_nick = nickserv
    login_command = identify mypassword
    rooms = #testroom1:#testroom2

    [irc.geeknode.org]
    nickname = anothernick
    login_nick = C
    login_command = nick identify mypassword
    rooms = #testvroum



.. _official website: http://biboumi.louiz.org/

"""

import asyncio

from typing import Optional, Tuple, List, Any
from slixmpp import JID, InvalidJID

from poezio.plugin import BasePlugin
from poezio.decorators import command_args_parser
from poezio.core.structs import Completion
from poezio import tabs


class Plugin(BasePlugin):
    default_config = {
        'irc': {
            "initial_connect": True,
            "gateway": "irc.jabberfr.org",
        }
    }

    def init(self) -> None:
        if self.config.getbool('initial_connect'):
            self.core.register_task(
                self.initial_connect()
            )
        self.api.add_command(
            'irc_join',
            self.command_irc_join,
            usage='<room or server>',
            help=('Join <room> in the same server as the '
                  'current tab (if it is an IRC tab). Or '
                  'join all the preconfigured rooms in '
                  '<server> '),
            short='Join irc rooms more easily',
            completion=self.completion_irc_join)

        self.api.add_command(
            'irc_query',
            self.command_irc_query,
            usage='<nickname> [message]',
            help=('Open a private conversation with the '
                  'given <nickname>, on the current IRC '
                  'server.  Optionally immediately send '
                  'the given message. For example, if the '
                  'current tab is #foo%irc.example.com@'
                  'biboumi.example.com, doing `/irc_query '
                  'nick "hi there"` is equivalent to '
                  '`/message nick%irc.example.com@biboumi.'
                  'example.com "hi there"`'),
            short='Open a private conversation with an IRC user')

    async def join(self, gateway: str, server: JID) -> None:
        "Join irc rooms on a server"
        nick: str = self.config.get_by_tabname(
            'nickname', server, default='') or self.core.own_nick
        rooms: List[str] = self.config.get_by_tabname(
            'rooms', server, default='').split(':')
        joins = []
        for room in rooms:
            room = '{}%{}@{}/{}'.format(room, server, gateway, nick)
            joins.append(self.core.command.join(room))

        await asyncio.gather(*joins)

    async def initial_connect(self) -> None:
        gateway: str = self.config.getstr('gateway')
        sections: List[str] = self.config.sections()

        sections_jid = []
        for sect in sections:
            if sect == 'irc':
                continue
            try:
                sect_jid = JID(sect)
                if sect_jid != sect_jid.server:
                    self.api.information(f'Invalid server: {sect}', 'Warning')
                    continue
            except InvalidJID:
                self.api.information(f'Invalid server: {sect}', 'Warning')
                continue
            sections_jid.append(sect_jid)

        for section in sections_jid:
            room_suffix = '%{}@{}'.format(section, gateway)

            already_opened = False
            for tab in self.core.tabs:
                if isinstance(tab, tabs.MucTab) and tab.name.endswith(room_suffix) and tab.joined:
                    already_opened = True
                    break

            if not already_opened:
                await self.join(gateway, section)

    @command_args_parser.quoted(1, 1)
    async def command_irc_join(self, args: Optional[List[str]]) -> None:
        """
        /irc_join <room or server>
        """
        if not args:
            self.core.command.help('irc_join')
            return
        sections: List[str] = self.config.sections()
        if 'irc' in sections:
            sections.remove('irc')
        if args[0] in sections:
            try:
                section_jid = JID(args[0])
            except InvalidJID:
                self.api.information(f'Invalid address: {args[0]}', 'Error')
                return
            await self.join_server_rooms(section_jid)
        else:
            await self.join_room(args[0])

    @command_args_parser.quoted(1, 1)
    def command_irc_query(self, args: Optional[List[str]]) -> None:
        """
        Open a private conversation with the given nickname, on the current IRC
        server.
        """
        if not args:
            self.core.command.help('irc_query')
            return
        current_tab_info = self.get_current_tab_irc_info()
        if not current_tab_info:
            return
        server, gateway = current_tab_info
        nickname = args[0]
        message = None
        if len(args) == 2:
            message = args[1]
        jid = '{}%{}@{}'.format(nickname, server, gateway)
        if message:
            self.core.register_task(
                self.core.command.message('{} "{}"'.format(jid, message))
            )
        else:
            self.core.register_task(
                self.core.command.message('{}'.format(jid))
            )

    async def join_server_rooms(self, section: JID) -> None:
        """
        Join all the rooms configured for a section
        (section = irc server)
        """
        gateway: str = self.config.getstr('gateway')
        rooms: List[str] = self.config.get_by_tabname('rooms', section).split(':')
        nick: str = self.config.get_by_tabname('nickname', section)
        if nick:
            nick = '/' + nick
        else:
            nick = ''
        suffix = '%{}@{}{}'.format(section, gateway, nick)

        for room in rooms:
            await self.core.command.join(room + suffix)

    async def join_room(self, name: str) -> None:
        """
        Join a room with only its name and the current tab
        """
        current_tab_info = self.get_current_tab_irc_info()
        if not current_tab_info:
            return
        server, gateway = current_tab_info
        try:
            server_jid = JID(server)
        except InvalidJID:
            return

        room = '{}%{}@{}'.format(name, server, gateway)
        if self.config.get_by_tabname('nickname', JID(server_jid.bare)):
            room += '/' + self.config.get_by_tabname('nickname', JID(server_jid.bare))

        await self.core.command.join(room)

    def get_current_tab_irc_info(self) -> Optional[Tuple[str, str]]:
        """
        Return a tuple with the irc server and the gateway hostnames of the
        current tab. If the current tab is not an IRC channel or private
        conversation, a warning is displayed and None is returned
        """
        gateway: str = self.config.getstr('gateway')
        current = self.api.current_tab()
        if not isinstance(current, tabs.ChatTab):
            return None
        current_jid = current.jid
        if not current_jid.server == gateway:
            self.api.information(
                'The current tab does not appear to be an IRC one', 'Warning')
            return None
        if isinstance(current, tabs.OneToOneTab):
            if '%' not in current_jid.node:
                server = current_jid.node
            else:
                ignored, server = current_jid.node.rsplit('%', 1)
        elif isinstance(current, tabs.MucTab):
            if '%' not in current_jid.node:
                server = current_jid.node
            else:
                ignored, server = current_jid.node.rsplit('%', 1)
        else:
            self.api.information(
                'The current tab does not appear to be an IRC one', 'Warning')
            return None
        return server, gateway

    def completion_irc_join(self, the_input: Any) -> Completion:
        """
        completion for /irc_join
        """
        sections: List[str] = self.config.sections()
        if 'irc' in sections:
            sections.remove('irc')
        return Completion(the_input.new_completion, sections, 1)
